﻿using OpenCvSharp;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ConsoleApp1
{
    class Program
    {
        static void Main(string[] args)
        {
            Action<string, Func<Mat, Window, int>, int, Mat, Window> action = (title, work, times, mat, window) =>
            {
                Stopwatch st = new Stopwatch();
                st.Start();

                for (int i = 0; i < times; i++)
                {
                    work(mat, window);
                }

                st.Stop();

                Console.WriteLine(title + " : " + st.ElapsedMilliseconds);
            };

            List<Window> showList = new List<Window>();
            bool doShow = false;

            for (int i = 0; i < 5; i++)
            {
                showList.Add((doShow == true) ? new Window() : null);
            }

            using (Mat mat = new Mat(new Size(1920 * 3, 1080), MatType.CV_8UC3))
            {
                mat.SetTo(Scalar.Aqua);

                int count = 100;
                action($"(1) - jit", Convert, 1, mat, showList[0]);
                action($"opencv({count})", Convert, count, mat, null);

                action($"(1) - jit", Convert2, 1, mat, showList[1]);
                action($"c# for({count})", Convert2, count, mat, null);

                action($"(1) - jit", Convert3, 1, mat, showList[2]);
                action($"c# parallel for({count})", Convert3, count, mat, null);

                action($"(1) - jit", Convert4, 1, mat, showList[3]);
                action($"C++ for({count})", Convert4, count, mat, null);

                action($"(1) - jit", Convert5, 1, mat, showList[4]);
                action($"C++ parallel for({count})", Convert5, count, mat, null);
            }

            Application.Run();
        }

        static int Convert(Mat mat, Window window)
        {
            using (Mat dstMat = mat.CvtColor(ColorConversionCodes.BGR2BGRA))
            {
                if (window != null)
                {
                    window.ShowImage(dstMat);
                }

                return dstMat.Width;
            }
        }

        static unsafe int Convert2(Mat srcMat, Window window)
        {
            byte *srcPtr = (byte *)srcMat.Data;

            using (Mat dstMat = new Mat(new Size(1920 * 3, 1080), MatType.CV_8UC4))
            {
                byte *dstPtr = (byte *)dstMat.Data;

                for (int y = 0; y < srcMat.Height; y++)
                {
                    for (int x = 0; x < srcMat.Width - 1; x++)
                    {
                        int* src = (int*)srcPtr;
                        int* dst = (int*)dstPtr;

                        *dst = *src;
                        *(dstPtr + 3) = 0xff;

                        srcPtr = srcPtr + 3;
                        dstPtr = dstPtr + 4;
                    }

                    Buffer.MemoryCopy(srcPtr, dstPtr, 3, 3);
                    *(dstPtr + 3) = 0xff;
                }

                if (window != null)
                {
                    window.ShowImage(dstMat);
                }
            }

            return 0;
        }

        static unsafe int Convert3(Mat srcMat, Window window)
        {
            byte* srcPtr = (byte*)srcMat.Data;

            using (Mat dstMat = new Mat(new Size(1920 * 3, 1080), MatType.CV_8UC4))
            {
                byte* dstPtr = (byte*)dstMat.Data;

                Parallel.For(0, srcMat.Height, (y) =>
                {
                    byte* srcPtrY = srcPtr + (y * srcMat.Width * 3);
                    byte* dstPtrY = dstPtr + (y * srcMat.Width * 4);

                    for (int x = 0; x < srcMat.Width - 1; x++)
                    {
                        int* src = (int*)srcPtrY;
                        int* dst = (int*)dstPtrY;

                        *dst = *src;
                        *(dstPtrY + 3) = 0xff;

                        srcPtrY = srcPtrY + 3;
                        dstPtrY = dstPtrY + 4;
                    }

                    Buffer.MemoryCopy(srcPtrY, dstPtrY, 3, 3);
                    *(dstPtrY + 3) = 0xff;

                    // 근소한 차이로 위의 코드가 빠름.
                    //for (int x = 0; x < srcMat.Width; x++)
                    //{
                    //    *(dstPtrY + 0) = *(srcPtrY + 0);
                    //    *(dstPtrY + 1) = *(srcPtrY + 1);
                    //    *(dstPtrY + 2) = *(srcPtrY + 2);

                    //    *(dstPtrY + 3) = 0xff;

                    //    srcPtrY = srcPtrY + 3;
                    //    dstPtrY = dstPtrY + 4;
                    //}
                });

                if (window != null)
                {
                    window.ShowImage(dstMat);
                }
            }

            return 0;
        }


        [DllImport("MemCopyLib.dll")]
        public unsafe extern static void RGB2RGBA(byte* srcPtr, byte* dstPtr, int width, int height);

        [DllImport("MemCopyLib.dll")]
        public unsafe extern static void RGB2RGBA_Parallel(byte* srcPtr, byte* dstPtr, int width, int height);

        static unsafe int Convert4(Mat srcMat, Window window)
        {
            byte* srcPtr = (byte*)srcMat.Data;
            using (Mat dstMat = new Mat(new Size(1920 * 3, 1080), MatType.CV_8UC4))
            {
                byte* dstPtr = (byte*)dstMat.Data;
                RGB2RGBA(srcPtr, dstPtr, srcMat.Width, srcMat.Height);

                if (window != null)
                {
                    window.ShowImage(dstMat);
                }
            }

            return 0;
        }

        static unsafe int Convert5(Mat srcMat, Window window)
        {
            byte* srcPtr = (byte*)srcMat.Data;
            using (Mat dstMat = new Mat(new Size(1920 * 3, 1080), MatType.CV_8UC4))
            {
                byte* dstPtr = (byte*)dstMat.Data;
                RGB2RGBA_Parallel(srcPtr, dstPtr, srcMat.Width, srcMat.Height);

                if (window != null)
                {
                    window.ShowImage(dstMat);
                }
            }

            return 0;
        }
    }
}
